import hashlib
import os
import os.path
import tempfile

import batou.component
import batou.lib.cron
import batou.lib.download
import batou.lib.file
import batou.lib.nagios
import pkg_resources
import six


class Certificate(batou.component.Component):
    """SSL certificate management using let's encrypt -- or not

    Usage::

        # Add certificate component. After this step, a key and certificate
        # is available. In case of Let's Encrypt it's a self-signed one.
        self.cert = Certificate(
            self.public_name,
            docroot=self.docroot,
            key_content=self.key_content,
            crt_content=self.crt_content,
            use_letsencrypt=self.letsencrypt,
            extracommand='sudo systemctl reload nginx',
        )
        self += self.cert

        # Configure web server. Use `key` and `fullchain` attributes to get
        # the paths to the key and the certificate:
        self += batou.lib.file.File(
            '/etc/local/nginx/myconfig.conf')

        # Activate the configuration.
        self += batou_ext.nix.Rebuild()

        # Actually get a proper Let's Encrypt certificate. This step does
        # nothing, if you don't use LE.
        self += self.cert.activate_letsencrypt()


    If you need to run some command after executing dehydrated, e.g.
    restarting a service, you can use the extracommand argument
    to configure it. It will be called everytime the script is invoked.

    The component is setting up a cronjob (in case of you are using
    Let's encrypt). So you need to add the CronTab-component to your
    deployment. Import::

        from batou.lib.cron import CronTab

    ... and add the crontab component to the nginx host in your environment.

    """

    _required_params_ = {
        "wellknown": "hello",
    }

    # Let's Encrypt, dehydrated v0.7.0
    dehydrated_url = (
        "https://raw.githubusercontent.com/dehydrated-io/dehydrated"
        "/082da2527cb4aaa3a4740ba03e550205b076f822/dehydrated"
    )
    dehydrated_checksum = "md5:95a90950d3b9c01174e4f4f98cf3bd53"
    dehydrated_publickey_algo = batou.component.Attribute(
        str, default="secp384r1"
    )

    extracommand = None

    namevar = "domain"
    domain = None
    alternative_names = ()

    wellknown = None
    docroot = None

    refresh_timing = None

    # Optinal if you are having a valid certificate and don't want to
    # make usage of letsencrypt
    key_content = None
    crt_content = None

    # OSCP stapling
    trusted_crt_content = None

    use_letsencrypt = batou.component.Attribute("literal", default=True)

    letsencrypt_ca = "https://acme-v02.api.letsencrypt.org/directory"
    letsencrypt_challenge = "http-01"
    letsencrypt_hook = ""
    # Allow selecting an alternative root chain,
    # e.g. "ISRG Root X1" for the short chain after the X3 expiry.
    letsencrypt_alternative_chain = None

    # Whether a certificate check should be deployed, too
    # You will need something like nrpehost or sensuchecks on the host
    enable_check = batou.component.Attribute("literal", default=True)

    granted_user = batou.component.Attribute(str, default="nginx")

    _may_need_to_generate_certificates = False

    def configure(self):
        if not isinstance(self.alternative_names, (tuple, list)):
            raise ValueError(
                '"alternative_names" needs to be a tuple of string.'
            )
        if not self.refresh_timing:
            h = int(
                hashlib.md5(
                    six.ensure_binary(self.domain, "UTF-8")
                ).hexdigest(),
                16,
            )
            self.refresh_timing = "{} {} * * *".format(h % 60, h % 24)
        if self.key_content and not self.use_letsencrypt:
            self.crt_file = batou.lib.file.File(
                os.path.join("{}/{}.crt".format(self.workdir, self.domain)),
                content=self.crt_content,
            )
            self += self.crt_file
            self.key_file = batou.lib.file.File(
                os.path.join("{}/{}.key".format(self.workdir, self.domain)),
                content=self.key_content,
                mode=0o600,
                sensitive_data=True,
            )
            self += self.key_file

            if self.trusted_crt_content:
                self.trusted_file = batou.lib.file.File(
                    "{}/{}.trust.crt".format(self.workdir, self.domain),
                    content=self.trusted_crt_content,
                    mode=0o600,
                )
                self += self.trusted_file
                self.trusted = self.trusted_file.path

            self.key = self.key_file.path
            self.fullchain = self.crt_file.path

        else:
            self._may_need_to_generate_certificates = True
            self.key_dir = os.path.join(self.workdir, self.domain)
            self.key = os.path.join(self.key_dir, "privkey.pem")
            self.fullchain = os.path.join(self.key_dir, "fullchain.pem")

        if self.use_letsencrypt:
            # Okay, let's encrypt it is. There are two situations:
            # 1. bootstrap -- there is nothing.
            # 2. there already is a cert, either replace it with
            #    LE or refresh existing LE.

            self += batou.lib.download.Download(
                self.dehydrated_url,
                checksum=self.dehydrated_checksum,
                target="dehydrated",
            )
            self += batou.lib.file.Mode("dehydrated", mode=0o755)

            if not self.wellknown and self.docroot:
                self.wellknown = "{}/.well-known/acme-challenge".format(
                    self.docroot
                )
            self += batou.lib.file.File(
                self.wellknown, ensure="directory", leading=True
            )

            self += batou.lib.file.File(
                self.expand("domains-{{component.domain}}.txt"),
                content=self.expand(
                    "{{component.domain}} {{alternative}}",
                    alternative=" ".join(self.alternative_names),
                ),
            )
            self.domains_txt = self._

            self += batou.lib.file.File(
                "cert-{}.conf".format(self.domain),
                content=self.expand(
                    """
WELLKNOWN={{component.wellknown}}
CA={{component.letsencrypt_ca}}
CHALLENGETYPE={{component.letsencrypt_challenge}}
HOOK={{component.letsencrypt_hook}}
DOMAINS_TXT={{component.domains_txt.path}}
KEY_ALGO={{component.dehydrated_publickey_algo}}
"""
                ),
            )
            self.config = self._

            self += batou.lib.file.File(
                self.expand("cert-{{component.domain}}.sh"),
                content=pkg_resources.resource_string(
                    "batou_ext", "resources/cert.sh"
                ),
                mode=0o700,
            )
            self.cert_sh = self._

            self += batou.lib.cron.CronJob(
                self.cert_sh.path,
                timing=self.refresh_timing,
                logger="cert-update",
            )

        if self.enable_check:
            self += CertificateCheck(self.domain)

    def activate_letsencrypt(self):
        """Return a component which really activates LE"""
        return ActivateLetsEncrypt(cert=self)

    def verify(self):
        if not self._may_need_to_generate_certificates:
            return
        if os.path.exists(self.key) and os.path.exists(self.fullchain):
            # So there are certificates. All done.
            return
        raise batou.UpdateNeeded()

    def update(self):
        # Create a temporary, self-signed certificate, to let the web server
        # start up, so let's encrypt can do what it needs.
        if not os.path.isdir(self.key_dir):
            os.makedirs(self.key_dir)
        self.csr_file = tempfile.NamedTemporaryFile()
        self.cmd("openssl genrsa -out {{component.key}} 2048")
        self.cmd(
            """\
openssl req -new \
    -key {{component.key}} \
    -out {{component.csr_file.name}} \
    -batch \
    -subj "/CN={{component.domain}}/emailAddress=admin@{{component.domain}}/C=DE"
"""
        )  # noqa
        self.cmd(
            """
openssl x509 -req -days 3650 \
    -in {{component.csr_file.name}} \
    -signkey {{component.key}} \
    -out {{component.fullchain}}
"""
        )
        self.cert_dir = os.path.join(self.workdir, self.domain)
        self.cmd(f"setfacl -Rm u:{self.granted_user}:rX {self.cert_dir}")
        self.csr_file.close()
        del self.csr_file


class ActivateLetsEncrypt(batou.component.Component):

    cert: Certificate = None

    def verify(self):
        if self.cert.use_letsencrypt:
            self.cert.assert_no_subcomponent_changes()

    def update(self):
        # If batou thinks we need to run for a new certificate because
        # we changed settings etc. we need to speed up certificate renewal.
        self.cmd(self.cert.cert_sh.path + " -x")

    @property
    def namevar_for_breadcrumb(self):
        return self.cert.namevar_for_breadcrumb


class CertificateCheck(batou.component.Component):

    namevar = "public_name"
    warning_days = 25
    critical_days = 14

    # If HTTPS is not running on 443
    port = batou.component.Attribute(int, default=443)

    def configure(self):
        self += batou.lib.nagios.ServiceCheck(
            self.expand("https://{{component.public_name}} certificate valid?"),
            name=self.expand("ssl_cert_{{component.public_name}}"),
            command="check_http",
            args=self.expand(
                "-H {{component.public_name}} "
                "-p {{component.port}} "
                "-S --sni "
                "-C {{component.warning_days}},{{component.critical_days}}"
            ),
        )


class CertificateCheckLocal(batou.component.Component):
    """
    This component helps to check whether a local certificate has expired
    or will expire soon. Useful in e.g. combination with
    client-certificates.

    Usage:

    self += batou_ext.ssl.CertificateCheckLocal(
        'path/to/your/certificate',
        name='My client certificate',
        warning_days=30,
        critical_days=10)

    """

    _required_params_ = {
        "name": "cert",
    }
    namevar = "certificate_path"
    name = None
    warning_days = 25
    critical_days = 14

    def configure(self):
        self.critical_seconds = self.critical_days * 24 * 3600
        self.warning_seconds = self.warning_days * 24 * 3600

        if self.name is None:
            raise ValueError("Required name is missing from certificate check")

        self += batou.lib.file.File(
            "cert_check_{}.sh".format(self.name),
            content=pkg_resources.resource_string(
                __name__, "resources/ssl/local_certificate_check.sh"
            ),
            mode=0o755,
        )
        self.script = self._.path
        self += batou.lib.nagios.ServiceCheck(
            self.expand("{{component.certificate_path}} certificate valid?"),
            name=self.name,
            command=self.script,
        )



import os

from batou import UpdateNeeded
from batou.component import Attribute, Component, ConfigString
from batou.lib.archive import Extract
from batou.lib.download import Download
from batou.lib.file import Directory, File, SyncDirectory
from batou.utils import Address

from batou_ext.php import FPM


class Roundcube(Component):
    """Configure Roundcube with database connection.

    Roundcube is installed with php/fastcgi. A basic configuration for the
    frontend is created but it is up to the user to fine-tune that
    configuration.
    """

    release = "1.1.4"
    checksum = "sha256:9bfe88255d4ffc288f5776de1cead78352469b1766d5ebaebe6e28043affe181"  # noqa: E501 line too long

    address = Attribute(Address, ConfigString("127.0.0.1:9000"))
    skin = "larry"
    support_url = "http://localhost"

    smtp_user = "%u"
    smtp_pass = "%p"

    config = os.path.join(os.path.dirname(__file__), "config.inc.php")

    def configure(self):
        self.db = self.require_one("roundcube::database")
        postfix = self.require_one("postfix")

        self.imap_host = postfix.connect.host
        self.smtp_server = postfix.connect.host
        self.smtp_port = postfix.connect.port

        self.basedir = self.map("roundcube")
        self.provide("roundcube", self)

        self += Directory("download")
        download = Download(
            "http://downloads.sourceforge.net/project/roundcubemail/"
            "roundcubemail/{}/roundcubemail-{}-complete.tar.gz".format(
                self.release, self.release
            ),
            target="download/roundcube-{}.tar.gz".format(self.release),
            checksum=self.checksum,
        )
        self += download

        self += Extract(download.target, target="roundcube.orig")
        self += SyncDirectory(
            self.basedir,
            source=self.map(
                "roundcube.orig/roundcubemail-{}".format(self.release)
            ),
        )

        self.db_dsnw = "{}://{}:{}@{}/{}".format(
            self.db.dbms,
            self.db.username,
            self.db.password,
            self.db.address.connect.host,
            self.db.database,
        )

        self += File(
            self.basedir + "/config/config.inc.php", source=self.config
        )

        self.fpm = FPM("roundcube")
        self += self.fpm

        self += RoundcubeInit(self)


class RoundcubeInit(Component):

    namevar = "roundcube"

    def verify(self):
        os.environ["PGPASSWORD"] = self.roundcube.db.password
        host = self.roundcube.db.address.connect.host
        db = self.roundcube.db
        try:

            result = self.cmd(
                f"psql -h {host} -U {db.username} -d {db.database} -c "
                '"SELECT * FROM information_schema.tables '
                "WHERE table_schema='public' "
                "AND table_catalog='roundcube';\""
            )
            if "contact" not in result[0]:
                raise UpdateNeeded()
        finally:
            del os.environ["PGPASSWORD"]

    def update(self):
        os.environ["PGPASSWORD"] = self.roundcube.db.password
        self.cmd(
            "psql -h {} -d {} -U {} -f {}/SQL/postgres.initial.sql".format(
                self.roundcube.db.address.connect.host,
                self.roundcube.db.database,
                self.roundcube.db.username,
                self.roundcube.basedir,
            )
        )
        del os.environ["PGPASSWORD"]



import hashlib
import os

from batou.component import Attribute, Component, ConfigString
from batou.lib.archive import Extract
from batou.lib.download import Download
from batou.lib.file import File, SyncDirectory
from batou.utils import Address

from batou_ext.php import FPM


class PFA(Component):

    _required_params_ = {
        "admin_password": "tiger",
    }
    release = "2.92"
    checksum = "sha1:21481f6eb8f10ba05fc6fcd1fe0fd468062956f2"

    address = Attribute(Address, ConfigString("127.0.0.1:9001"))

    admin_password = None
    salt = "ab8f1b639d31875b59fa047481c581fd"
    config = os.path.join(
        os.path.dirname(__file__), "postfixadmin", "config.local.php"
    )

    def configure(self):
        self.db = self.require_one("pfa::database")
        self.postfix = self.require_one("postfix")
        self.provide("pfa", self)

        self.basedir = self.map("postfixadmin")

        download = Download(
            "http://downloads.sourceforge.net/project/postfixadmin/"
            "postfixadmin/postfixadmin-{}/postfixadmin-{}.tar.gz".format(
                self.release, self.release
            ),
            target="postfixadmin-{}.tar.gz".format(self.release),
            checksum=self.checksum,
        )
        self += download
        self += Extract(download.target, target="postfixadmin.orig")

        self += SyncDirectory(
            self.basedir,
            source=self.map(
                "postfixadmin.orig/postfixadmin-{}".format(self.release)
            ),
        )

        self += File(self.basedir + "/config.local.php", source=self.config)

        self.fpm = FPM("postfixadmin")
        self += self.fpm

    @property
    def admin_password_encrypted(self):
        # password generation ported from postfixadmin/setup.php
        encrypt = hashlib.sha1()
        encrypt.update(
            "{}:{}".format(self.salt, self.admin_password).encode("utf-8")
        )

        return "{}:{}".format(self.salt, encrypt.hexdigest())



import copy
import os.path
import re

import batou.component


def dict_merge(a, b):
    """recursively merges dict's. not just simple a['key'] = b['key'], if
    both a and b have a key who's value is a dict then dict_merge is called
    on both values and the result stored in the returned dictionary.
    https://www.xormedia.com/recursively-merge-dictionaries-in-python/
    """
    if not isinstance(b, dict):
        return b
    result = copy.deepcopy(a)
    for k, v in b.items():
        if k in result and isinstance(result[k], dict):
            result[k] = dict_merge(result[k], v)
        elif k in result and isinstance(result[k], list):
            result[k] = result[k][:]
            result[k].extend(v)
        else:
            result[k] = copy.deepcopy(v)
    return result


class RegexPatch(batou.component.Component):
    r"""Patch existing file with a regexp.

    Usage::

        self += batou_ext.config.RegexPatch(
            '/path/to/file',
            pattern=r'^foo=\d+(\w+)',
            replacement=r'foo=27\1')

    If you don't want to patch inplace:

        self += batou_ext.config.RegexPatch(
            '/path/to/file',
            source='/path/to/some/file/'
            pattern=r'^foo=\d+(\w+)',
            replacement=r'foo=27\1')

    """

    _required_params_ = {
        "pattern": "pattern",
    }
    path = None
    namevar = "path"

    source = None
    pattern = None
    replacement = None

    def configure(self):
        self.pattern = re.compile(self.pattern, re.MULTILINE)

        if not self.path:
            raise ValueError("`path` must be set.")

        if not self.source:
            self.source = self.path

    def _patch(self):
        return self.pattern.sub(self.replacement, self._source_data)

    def verify(self):

        if not os.path.exists(self.source):
            # During predict, the file might not exist.
            raise batou.UpdateNeeded()
        with open(self.source, "r") as f:
            self._source_data = f.read()

        if not os.path.exists(self.path):
            # During predict, the file might not exist.
            raise batou.UpdateNeeded()
        with open(self.path, "r") as f:
            self._target_data = f.read()

        m = self.pattern.search(self._source_data)
        assert m, "Could not configure, no match for pattern: {}".format(
            self.pattern.pattern
        )
        if self._target_data != self._patch():
            raise batou.UpdateNeeded()

    def update(self):
        if not os.path.exists(self.source):
            raise IOError("The file to be patched does not exist:", self.source)

        if self._source_data is None:
            raise RuntimeError(
                "The file to be patched seems to have no valid content:",
                self.source,
            )

        with open(self.path, "w") as f:
            f.write(self._patch())


class MultiRegexPatch(batou.component.Component):
    """Patch existing file with multiple regexp.

    Usage::

        self += batou_ext.config.MultiRegexPatch(
            '/path/to/file', patterns=[
            (r'pattern-to-match',
             r'value-to-replace-with'),

            (r'another-pattern',
             r'another-replacement'),
        ])

    """

    namevar = "path"
    patterns = ()

    def configure(self):
        for pattern, replacement in self.patterns:
            self += RegexPatch(
                self.path,
                pattern=pattern,
                replacement=self.parent.expand(replacement),
            )



import os.path

import batou.component
import batou.lib.file

import batou_ext.cron


class GeoIPDatabase(batou.component.Component):
    """
    A small component to download a GeoIP-database and keep downloaded
    up to date.

    Usage

    self += batou_ext.geoip.GeoIPDatabase()
    self.geoipdb = self._.database_file

    Needs curl and gunzip inside $PATH
    """

    license_key = None
    download_url = batou.component.Attribute(
        str,
        batou.component.ConfigString(
            "https://download.maxmind.com/app/geoip_download?"
            "edition_id=GeoLite2-City&suffix=tar.gz&"
            "license_key={{component.license_key}}"
        ),
    )

    def configure(self):

        self.provide("geoip_database", self)
        self += batou.lib.file.File(
            "geoip-update.sh",
            source=os.path.join(
                os.path.dirname(__file__), "resources/geoip-update.sh"
            ),
            mode=0o744,
        )
        self.script = self._.path

        self += batou_ext.cron.CronJob(
            "GeoIP_update",
            command=self.script,
            timing="@weekly",
            timeout="1h",
            checkWarning=15000,
            checkCritical=30000,
        )

        self.database_file = self.expand(
            "{{component.workdir}}/GeoLite2-City.mmdb"
        )

    def verify(self):
        self.assert_no_changes()

    def update(self):
        self.cmd("{{component.workdir}}/geoip-update.sh", expand=True)



import os.path

import batou.lib.archive


class SingleUntar(batou.lib.archive.Untar):
    """Extract archive *once*

    This is useful when the target dir contains a version number or revision.
    In this case we do not expect that extracting the archive again could do
    any changes, so we skip the file-by-file validation which speeds things
    up considerably.

    """

    _required_params_ = {"target": "."}

    def configure(self):
        super().configure()
        self._finish_marker = os.path.join(
            os.path.dirname(self.target),
            f".{os.path.basename(self.target)}.finished-extract",
        )

    def verify(self):
        assert os.path.exists(self.archive)
        # Verify that the target dir is newer than the archive on disk.
        self.assert_file_is_current(self.target, [self.archive])
        # Verify that there is a finish marker and it is newer than the arvive
        self.assert_file_is_current(self._finish_marker, [self.archive])

    def update(self):
        super().update()
        self.touch(self._finish_marker)



import glob
import os
import os.path
import shutil
import subprocess
import urllib.parse

import batou
import batou.component
import batou.lib.file
import batou.lib.git

import batou_ext.ssh


class SymlinkAndCleanup(batou.component.Component):
    """
    Symlink the give file or directory to `current`, symlink the file or directory that was symlinked to `current` to `last` and remove all files matching `pattern` that are not currently linked to current or last.
    This utility is mostly used for Git Checkouts, which can grow considerably in size, but may also be used for S3 File Downloads or other Folders and Files that should be cached.
    """

    namevar = "current"
    pattern = None
    _current_link = None
    _last_link = None

    prefix = None

    use_systemd_run_async_cleanup = False

    # Use extra args to e.g. limit IOPS:
    #     ["--property=IOReadIOPSMax=/dev/vda 100",
    #      "--property=IOWriteIOPSMax=/dev/vda 100"]
    systemd_extra_args: list = None

    def configure(self):
        self._current_link = (
            f"{self.prefix}-current" if self.prefix else "current"
        )
        self._last_link = f"{self.prefix}-last" if self.prefix else "last"
        self.dir = os.path.dirname(self.current)
        self.current = os.path.basename(self.current)

    @staticmethod
    def _link(path):
        try:
            return os.readlink(path)
        except OSError:
            batou.output.error(f"Unable to read link: `{path}`")
            return None

    @staticmethod
    def remove(links, el):
        try:
            links.remove(el)
        except ValueError:
            pass

    def _list_removals(self):
        candidates = glob.glob(self.pattern)

        current = self._link(self._current_link)
        last = self._link(self._last_link)
        if current == self.current:
            # keep last+current
            self.remove(candidates, current)
            self.remove(candidates, last)
        else:
            # keep current + new current"
            self.remove(candidates, current)
            self.remove(candidates, self.current)

        return candidates

    def verify(self):
        with self.chdir(self.dir):
            if self._link(self._current_link) != self.current:
                raise batou.UpdateNeeded()
            if self._list_removals():
                raise batou.UpdateNeeded()

    def update(self):
        with self.chdir(self.dir):
            current = self._link(self._current_link)
            if current != self.current:
                try:
                    os.remove(self._current_link)
                except OSError as e:
                    batou.output.error(
                        f'Failed to remove link "{self._current_link}" ({current}): {e.strerror}'
                    )
                    pass

                try:
                    os.remove(self._last_link)
                except OSError as e:
                    last = self._link(self._last_link)
                    batou.output.error(
                        f'Failed to remove link "{self._last_link}" ({last}): {e.strerror}'
                    )
                    pass

                batou.output.annotate("current -> {}".format(self.current))
                os.symlink(self.current, self._current_link)
                if current:
                    batou.output.annotate("last -> {}".format(current))
                    os.symlink(current, self._last_link)

            candidates = self._list_removals()
            if not candidates:
                batou.output.annotate("Nothing to remove.")
            elif self.use_systemd_run_async_cleanup:
                # Spawns an systemd-run cleanup job with custom args.
                candidates = [os.path.join(self.dir, c) for c in candidates]
                rm_cmd = [
                    "rm",
                    "-rfv",
                    *candidates,  # consider: ARG_MAX is limited
                ]
                extra_args = self.systemd_extra_args or []
                cmd = [
                    "systemd-run",
                    "--unit",
                    f"batou-cleanup-{self.prefix}",
                    "--user",
                    *extra_args,
                    *rm_cmd,
                ]
                batou.output.annotate(f"Removing: {candidates}")
                batou.output.annotate(f"    {cmd}")
                try:
                    subprocess.run(cmd, check=True)
                except subprocess.CalledProcessError as e:
                    batou.output.error(f"Failed to remove: {e}")
            else:
                for el in candidates:
                    batou.output.annotate("Removing: {}".format(el))
                    try:
                        if os.path.isdir(el):
                            shutil.rmtree(el)
                        else:
                            os.remove(el)
                    except OSError as e:
                        batou.output.error(
                            f'Failed to remove "{el}": {e.strerror}'
                        )



import os
import shlex
from textwrap import dedent
from typing import Optional

import batou
from batou import UpdateNeeded
from batou.component import Attribute, Component
from batou.lib.file import File
from batou.utils import CmdExecutionError

import batou_ext.nix


@batou_ext.nix.rebuild
class Container(Component):
    """A OCI Container component.

    With this component you can dynamically schedule docker containers to be
    run on the target host.

    Note: Docker image specifiers do not follow a properly resolvable pattern.
    Therefore, container registries have to be specified seperately if you need
    to log in before. If you do not provide a container registry, docker will
    use the default one for authentication. You can choose to also append the
    image attribute with the registry but this module will do so automatically.

    ```
    # the following two call are identical:
    self += batou_ext.oci.Container(
        image="foo",
        registry_address="test.registry",
        registry_user="foo",
        registry_password="bar")

    self += batou_ext.oci.Container(
        image="test.registry/foo",
        registry_address="test.registry",
        registry_user="foo",
        registry_password="bar")

    # However, this will fail since docker will try to log into the default container registry
    # with the provided credentials and then pull the image from the registry specified in the image string
    self += batou_ext.oci.Container(
        image="test.registry/foo",
        registry_user="foo",
        registry_password="bar")


    # if you don't need to authenticate, not explicitly specifying the registry is fine of course
    # and it will pull the image from the correct registry
    self += batou_ext.oci.Container(image="test.registry/foo")
    ```

    Example:
    ```
    self += batou_ext.oci.Container(image = "mysql", version = "8.0")
    ```

    """

    # general options
    image = Attribute(str)
    version: str = "latest"
    container_name: Optional[str] = None

    # specific options
    entrypoint: Optional[str] = None
    docker_cmd: Optional[str] = None
    envfile: Optional[File] = None
    mounts: dict = {}
    ports: dict = {}
    env: dict = {}
    depends_on: list = None
    extra_options: list = None

    # secrets
    registry_address = Attribute(Optional[str], None)
    registry_user = Attribute(Optional[str], None)
    registry_password = Attribute(Optional[str], None)

    _required_params_ = {
        "image": "mysql",
    }

    # cache spanning multiple components deploying the same container
    _remote_manifest_cache = {}

    def configure(self):
        if (
            self.registry_user or self.registry_password
        ) and not self.registry_address:
            self.log(
                "WARN: you might want to specify the registry explicitly"
                " unless you really intend to log into the default"
                " docker registry"
            )

        if (
            self.registry_address
            # This is for the strange case of index.docker.io where you have
            # to set the registry_address to `https://index.docker.io/v1/`
            and not self.registry_address.startswith("https://")
            and not self.image.startswith(self.registry_address)
        ):
            self.image = f"{self.registry_address}{'/' if not self.registry_address.endswith('/') else ''}{self.image}"

        if self.registry_password:
            self += File(
                "registry-password", mode=0o600, content=self.registry_password
            )
            self.password_file = self._

        if self.envfile is None:
            self += File(
                f"{self.container_name}_env",
                sensitive_data=True,
                content="""{% for key, value in component.env.items() | sort -%}
{{key}}={{value}}
{% endfor %}""",
            )
            self.envfile = self._

        if self.docker_cmd:
            self._docker_cmd_list = shlex.split(self.docker_cmd)

        if not self.depends_on:
            self.depends_on = []

        self += File(
            f"/etc/local/nixos/docker_{self.container_name}.nix",
            sensitive_data=False,
            source=os.path.join(
                os.path.dirname(__file__), "resources/oci-template.nix"
            ),
        )

    def verify(self):
        self.assert_no_changes()
        self.envfile.assert_no_changes()

        valid = False

        container_image_id, stderr = self.cmd(
            dedent(
                """\
            docker container insepct {{component.container_name}} \
                | jq -r '.[0].Image'
                """
            )
        )
        local_image_id, stderr = self.cmd(
            dedent(
                """\
            docker image inspect {{component.image}}:{{component.version}} \
                | jq -r '.[0].Id' \
                || echo image not available locally
                """
            )
        )
        if local_image_id != container_image_id:
            # If the container is not running the image we expect, we need to
            # restart it. If its the same, we need to dig further
            error = (
                "Container is running different image. "
                "({container_image_id} vs. {local_image_id})"
            )
            local_digest, stderr = self.cmd(
                dedent(
                    """\
                docker image inspect {{component.image}}:{{component.version}} \
                    | jq -r 'first | .RepoDigests | first | split("@") | last' \
                    || echo image not available locally
                    """
                )
            )

            image_ident = f"{self.image}:{self.version}@{local_digest}"
            try:
                valid = self._remote_manifest_cache[image_ident]
                error = "(cached)"
            except KeyError:
                if self.registry_address:
                    logintxt, _ = self.cmd(
                        self.expand(
                            dedent(
                                """\
                docker login \\
                    {%- if component.registry_user and component.registry_password %}
                    -u {{component.registry_user}} \\
                    -p {{component.registry_password}} \\
                    {%- endif %}
                    {{component.registry_address}}
                """
                            )
                        )
                    )

                try:
                    stdout, stderr = self.cmd(
                        f"docker manifest inspect {image_ident}"
                    )
                except CmdExecutionError as e:
                    valid = False
                    error = e.stderr
                    if error.startswith(
                        "unsupported manifest format"
                    ):  # gitlab
                        batou.output.annotate(error, debug=True)
                        error = error[:50]
                else:
                    # `docker manifest inspect` silently raises an error,
                    # returns code 0 when unathorized
                    if stderr == "unauthorized":
                        raise RuntimeError(
                            "Wrong credentials for remote container registry"
                        )
                    valid = True
                self._remote_manifest_cache[image_ident] = valid

        if not valid:
            self.log("Update due digest verification error: %r", error)
            raise UpdateNeeded()

    def update(self):
        self.cmd(f"sudo systemctl restart docker-{self.container_name}")



import os.path
import shlex
from glob import glob

import batou.component
import batou.lib.python
from batou.utils import CmdExecutionError


class Pipenv(batou.component.Component):
    """Sync pipenv

    Usage:

    * Checkout project to workdir, expecting Pipfile/Pipfile.lock there
    * self += Pipenv()`
    * Run services via `pipenv.executable` (points ot virualenv'ed python)

    Notes:

    * The Python version *is* specified in the Pipfile. So there is no need to
      pass.
    * `pipenv` is epxected to be there already. I.e. coming from the system.
    * For NixOS you can `nix-env -iA nixos.pipenv`. You will also need
      `nixos.which` and the python version the Pipfile specifies.

    """

    target = None

    def configure(self):
        if self.target is None:
            self.target = self.workdir
        self.venv = os.path.join(self.workdir, self.target, ".venv")
        self.executable = os.path.join(self.venv, "bin/python")

    def verify(self):
        with self.chdir(self.target):
            self.assert_file_is_current(
                self.executable, ["Pipfile", "Pipfile.lock"]
            )
            # Is this Python (still) functional 'enough'
            # from a setuptools/distribute perspective?
            self.assert_cmd(
                '{{component.executable}} -c "import pkg_resources"'
            )

    def update(self):
        with self.chdir(self.target):
            self.cmd("rm -rf .venv")
            self.cmd("pipenv sync", env={"PIPENV_VENV_IN_PROJECT": "1"})


class VirtualEnvRequirements(batou.component.Component):
    """
    Installs a Python VirtualEnv with a given requirements.txt

    Usage::
        self += VirtualEnvRequirements(
            version='2.7',
            requirements_path='/path/to/my/requirements.txt')
    """

    version = batou.component.Attribute(str, default="2.7")
    requirements_path = batou.component.Attribute(
        str, batou.component.ConfigString("requirements.txt")
    )

    # Shell script to be sourced before creating VirtualEnv and pip
    pre_run_script_path = None

    # Passing environmental variables to batou's cmd
    env = None

    # May pass pre-fabricated virtualenv
    venv = None

    def configure(self):

        if isinstance(self.requirements_path, str):
            self.requirements_paths = [self.requirements_path]
        elif isinstance(self.requirements_path, list):
            self.requirements_paths = self.requirements_path
        else:
            raise RuntimeError("Needs to be either string or list")

        if self.venv is None:
            self.venv = batou.lib.python.VirtualEnv(self.version)
        self += self.venv

    def verify(self):
        self.assert_no_changes()
        self.parent.assert_no_changes()

    def update(self):
        for req in self.requirements_paths:
            if self.pre_run_script_path:
                self.cmd(
                    (
                        "source {} && {} " "-m pip install --upgrade -r {}"
                    ).format(self.pre_run_script_path, self.venv.python, req),
                    env=self.env,
                )
            else:
                self.cmd(
                    "{} -m pip install --upgrade -r {}".format(
                        self.venv.python, req
                    ),
                    env=self.env,
                )


class FixELFRunPath(batou.component.Component):
    """
    Patches DT_RUNPATH & DT_RPATH of each shared object to either point to a
    given user env or a path relative to the shared object. Should only be used
    in conjunction with a virtual env with prebuilt binaries.

    Usage::
        self += FixELFRunPath(
            path="/path/to/venv",
            env_directory=os.path.expanduser('~/.nix-profile/lib')
        )

    Now each .so contains a DT_RPATH with

    * `~/.nix-profile/lib`: in this environment are external libraries of
      prebuilt binaries used by Python libraries. For instance a `libz.so`
      for `numpy`.

      Please keep in mind that each library in this environment will be used
      by the venv since DT_RPATH has the highest priority (over LD_LIBRARY_PATH
      and DT_RUNPATH).

    * `~/.nix-profile/lib/*` for each `*` that contains a shared object. This is
      necessary for e.g. libmysqlclient / libmariadb.so.3 which is in a subdir of
      `$out/lib`.

      This behavior can be turned off by setting `recurse_env_dir` to `False`.

    * `$ORIGIN/../foo.libs`. With this path, libraries relative to the
      shared library's path will be loaded. For instance, numpy
      installs `numpy.libs` with a few prebuilt libraries such as gfortran
      into a path relative its own libraries in the venv.

    Additional notes::

    * in the end it turned out to be necessary to use `DT_RPATH` instead of
      `DT_RUNPATH` for two reasons:

      * `numpy` among others expects `$ORIGIN/../../numpy.libs` to be in
        `DT_RPATH` since that is used for the library itself _and_ all other
        libraries below in the dependency tree (in contrast to DT_RUNPATH).

      * it's impossible to use both `DT_RPATH` and `DT_RUNPATH` for a single
        ELF binary since `DT_RPATH` will be ignored if `DT_RUNPATH` exists
        (see `ld.so(8)`).

      Despite the higher priority this is still preferable over
      `LD_LIBRARY_PATH` because it doesn't "infect" the entire process: the
      env gets inherited to child processes and causes side-effects there as
      well (see e.g. FC-35412, FC-36567). Also, DT_RPATH will only be used
      for the child dependencies of a binary. That means that it will be used
      for each library loaded by numpy's C libraries, but won't be used for e.g.
      psycopg2 and its libraries.

    * It's not desirable to use `glibc` inside an env used by this component:

      * It's still very easy to run into ABI issues: for instance for development
        I put a glibc <2.34 into the user env which still had a non-empty
        libpthread.so.0 (got merged into libc.so.6 in 2.34) which caused
        caused a conflict with how the newer glibc is structued.

      * Generally when using a recent Python and a recent e.g. numpy, there's
        a good chance it was linked against a relatively new glibc and
        providing an old glibc here will break.

      * When using psycopg2 (as opposed to psycopg2-binary), the library
        will be compiled on install. For that, it's better to install `gcc` into
        a user env to make sure that this is used as C compiler.
        Subsequently, it ensures that the correct glibc is provided on linking.
    """

    path = batou.component.Attribute(str)
    env_directory = batou.component.Attribute(str)
    glob_patterns = batou.component.Attribute(
        list, default=["**/*.so", "**/*.so.*"]
    )
    patchelf_jobs = batou.component.Attribute(int, default=4)
    recurse_env_dir = batou.component.Attribute(bool, default=True)

    def verify(self):
        self.assert_no_changes()
        self.parent.assert_no_changes()

    def update(self):
        if self.recurse_env_dir:
            with self.chdir(self.env_directory):
                directories = ":".join(
                    sorted(
                        {
                            self.env_directory + "/" + os.path.dirname(so_file)
                            for so_file in glob("**/*.so", recursive=True)
                        }
                    )
                )
        else:
            directories = self.env_directory

        with self.chdir(self.path):
            files_to_fix = [
                x
                for pattern in self.glob_patterns
                for x in glob(pattern, recursive=True)
            ]
            if not files_to_fix:
                return

            # add user env to DT_RPATH
            self.__patchelf(["--add-rpath", directories], files_to_fix)
            # drop everything from DT_RPATH except self.env_directory
            # and directories in the venv (to allow shared libraries from numpy
            # to load other shared libraries from numpy).
            self.__patchelf(
                [
                    "--shrink-rpath",
                    "--allowed-rpath-prefixes",
                    f"{directories}:$ORIGIN",
                ],
                files_to_fix,
            )

    def __patchelf(self, args, paths):
        args_ = " ".join(shlex.quote(arg) for arg in args)
        # `--force-rpath` because we need `rpath` for $ORIGIN since rpath
        # works for all ELFs below in the dependency tree in contrast to DT_RUNPATH.
        # It's impossible to use both at the same time because DT_RPATH will always
        # be ignored then. For more context, see `ld.so(8)`.
        patchelf = "nix run -f '<nixpkgs>' -- patchelf --force-rpath"
        cmd = f"xargs -P {self.patchelf_jobs} {patchelf} {args_}"
        proc = self.cmd(cmd, communicate=False)

        stdout, stderr = proc.communicate(
            input=bytes("\n".join(paths) + "\n", "utf-8")
        )
        if proc.returncode != 0:
            raise CmdExecutionError(cmd, proc.returncode, stdout, stderr)



import os

import batou
import batou.component
import batou.utils


class PostgresServer(batou.component.Component):

    listen_port = "5432"

    def configure(self):
        self.address = batou.utils.Address(self.host.fqdn, self.listen_port)
        self.provide("postgres", self)


class PostgresDataComponent(batou.component.Component):

    command_prefix = batou.component.Attribute(str, default="sudo -u postgres")

    def configure(self):
        try:
            self._command_prefix = self.parent.command_prefix
            self.log(
                "{}: Warning: The usage of command_prefix from parent-"
                "component will be deprecated in future. Please update "
                "your deployment to make usage of {}.command_prefix "
                "rather than defining it via {}.command_prefix.".format(
                    self._breadcrumbs,
                    self.__class__.__name__,
                    self.parent.__class__.__name__,
                )
            )
        except AttributeError:
            self._command_prefix = self.command_prefix

    def pgcmd(self, cmd, *args, **kw):
        return self.cmd("{} {}".format(self.command_prefix, cmd), *args, **kw)


class DB(PostgresDataComponent):
    """
    Ensures a given database is created and owned by a specific role.

    Usage:
    self += batou_ext.postgres.DB(
        "mydatabase",
        owner="myuser")

    Attention: The user will not be created automatically.
    """

    _required_params_ = {
        "owner": "scott",
    }
    namevar = "db"
    locale = "en_US.UTF-8"
    template = "template1"
    owner = None

    def configure(self):
        super(DB, self).configure()
        if self.owner is None:
            raise ValueError(
                "You have to specify an owner for the "
                'database "{}"'.format(self.db)
            )

    def verify(self):
        try:
            self.pgcmd(
                self.expand('psql -c "SELECT true;" -d "{{component.db}}"'),
                silent=True,
            )
        except batou.utils.CmdExecutionError:
            raise batou.UpdateNeeded()

    def update(self):
        self.pgcmd(
            self.expand(
                "createdb -T {{component.template}} -l {{component.locale}} -O"
                ' "{{component.owner}}" "{{component.db}}"'
            )
        )


class User(PostgresDataComponent):
    """
    Creates a user/role at the database cluster with a given password
    and flags.

    Usage:
    self += batou_ext.postgres.User(
        "crocodile",
        password="aligator3")
    """

    _required_params_ = {
        "password": "tiger",
    }
    namevar = "name"
    flags = "NOCREATEDB NOCREATEROLE NOSUPERUSER"
    password: str = None

    def configure(self):
        super(User, self).configure()
        assert self.password is not None, "Password for %s is None" % self.name

    def verify(self):
        os.environ["PGPASSWORD"] = self.password
        try:
            self.cmd(
                self.expand(
                    'psql -d postgres -c "SELECT true;" -U'
                    ' "{{component.name}}" -w -h localhost'
                )
            )
        except batou.utils.CmdExecutionError:
            raise batou.UpdateNeeded()
        finally:
            del os.environ["PGPASSWORD"]

    def update(self):
        try:
            command = self.expand(
                'psql -d postgres -c "CREATE USER \\"{{component.name}}\\" PASSWORD \'{{component.password}}\' {{component.flags}}"'  # noqa: E501 line too long
            )
            self.pgcmd(command)
        except batou.utils.CmdExecutionError as e:
            if "already exists" in e.stderr:
                command = self.expand(
                    'psql -d postgres -c "ALTER ROLE \\"{{component.name}}\\" WITH ENCRYPTED PASSWORD \'{{component.password}}\' "'  # noqa: E501 line too long
                )
                self.pgcmd(command)
            else:
                raise e


class Extension(PostgresDataComponent):
    """
    Creates an extension to a given PostgreSQL-DB.

    Usage:
    self += batou_ext.postgres.Extension(
        'uuid-ossp',
        db='my_database')

    Please note: The extension needs to be already present in context
    of the database cluster.
    """

    _required_params_ = {
        "db": "scott",
    }
    namevar = "extension_name"
    db = None

    def configure(self):
        if self.extension_name is None:
            raise ValueError("Need to set extension name")
        if self.db is None:
            raise ValueError(
                "Need to specify a database to " "create extension in"
            )

    def verify(self):
        cmd_out, cmt_err = self.pgcmd(
            self.expand(
                "psql -d {{component.db}} -qtAX "
                '-c "SELECT extname FROM pg_extension '
                "WHERE extname = '{{component.extension_name}}';\""
            )
        )
        if cmd_out.strip() != self.extension_name:
            raise batou.UpdateNeeded()

    def update(self):
        self.pgcmd(
            self.expand(
                "psql -c "
                '"CREATE EXTENSION '
                '\\"{{component.extension_name}}\\";" -d {{component.db}}'
            )
        )



import batou
import batou.component
import batou.lib.file
import passlib.handlers.sha2_crypt
import passlib.hash


class HTPasswd(batou.component.Component):

    _required_params_ = {"users": "user:pwd"}
    namevar = "path"
    users = ""

    def configure(self):
        users = []
        for line in self.users.split("\n"):
            line = line.strip()
            user, password = line.split(":")
            password = passlib.handlers.sha2_crypt.sha512_crypt.hash(password)
            users.append("{}:{}".format(user, password))
        self += batou.lib.file.File(self.path, content="\n".join(users))



"""A library of components for batou.
"""

import glob
import os.path

from setuptools import find_packages, setup


def project_path(*names):
    return os.path.join(*names)


setup(
    name="batou_ext",
    version="2.4.9",
    install_requires=[
        "batou >= 2.3b4",
        "pyaml",
        "setuptools",
        "six",
        "InquirerPy",
    ],
    extras_require={
        "test": [
            "boto3",
            "passlib>=1.7",
            "pytest",
            "pytest-mock",
        ],
        "version-select": ["InquirerPy"],
        "s3-bootstrap": ["boto3", "InquirerPy"],
    },
    author="Flying Circus <support@flyingcircus.io>",
    author_email="support@flyingcircus.io",
    license="BSD (2-clause)",
    url="https://github.com/flyingcircusio/batou_ext",
    keywords="deployment",
    classifiers="""\
License :: OSI Approved :: BSD License
Programming Language :: Python
Programming Language :: Python :: 3
Programming Language :: Python :: 3 :: Only
"""[
        :-1
    ].split(
        "\n"
    ),
    description=__doc__.strip(),
    long_description="\n\n".join(
        open(project_path(name)).read()
        for name in (
            "README.md",
            "CHANGES.md",
        )
    ),
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    data_files=[("", glob.glob(project_path("*.txt")))],
    entry_points=dict(
        console_scripts=[
            "jenkins = batou_ext.jenkins:main",
            "fcio = batou_ext.fcio:main",
            "versions = batou_ext.versions:main",
            "s3_bootstrap = batou_ext.s3_bootstrap:main [s3-bootstrap]",
        ]
    ),
    zip_safe=False,
)



#!/usr/bin/env python3

import subprocess
import sys

proc = subprocess.run(
    ["systemctl", "show", sys.argv[1], "-P", "ActiveState"],
    capture_output=True,
    text=True,
)

if proc.returncode != 0:
    print(f"ERROR: {proc.stderr}")
    sys.exit(2)

state = proc.stdout.strip()
if state == "failed":
    print(f"ERROR: {state}")
    sys.exit(2)

print(f"OK: {state}")
sys.exit(0)



import json
import shlex
import subprocess
from os import environ
from sys import exit
from typing import Dict, List, Union

import boto3
from InquirerPy import inquirer


class ANSIColors:
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"


def error(msg):
    print(ANSIColors.FAIL + str(msg) + ANSIColors.ENDC)


def print_bold(msg):
    print(ANSIColors.BOLD + str(msg) + ANSIColors.ENDC)


def remote_sudo(host: str, password: str, cmd: List[str]) -> str:
    cmd_ = " ".join(shlex.quote(x) for x in cmd)
    print_bold(f"[{host}]$ sudo -S {cmd_}")
    proc = subprocess.Popen(
        ["ssh", host, f"sudo -S {cmd_}"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    out, err = map(
        lambda x: x.decode("utf-8"),
        proc.communicate(input=bytes(f"{password}\n", "utf-8")),
    )

    exit_code = proc.wait()
    if exit_code != 0:
        raise Exception(
            f"Remote command {cmd} on {host} failed with exit code {exit_code}\nStdout:\n{out}\n\nStderr:{err}!"
        )

    return out


def rg_account_exists(name: str, host: str, sudo_password: str) -> bool:
    return name in json.loads(
        remote_sudo(host, sudo_password, ["radosgw-admin", "user", "list"])
    )


class Step:
    def __str__(self) -> str:
        return "Unknown step"

    def verify(self, previous_result):
        pass

    def apply(self, previous_result):
        pass


class Plan:
    steps: List[Step] = []

    def __add__(self, step: Step):
        self.steps.append(step)
        return self

    def __str__(self) -> str:
        return "\n".join(f" - {str(s)}" for s in self.steps)

    def apply(self):
        previous_result = None
        for step in self.steps:
            try:
                step.verify(previous_result)
            except StateAlreadyExists as e:
                error(e)
                print_bold("Aborting")
                exit(1)
            except ConfigurationError as e:
                error(e)
                print_bold("Aborting")
                exit(1)
            previous_result = step.apply(previous_result)


class StateAlreadyExists(Exception):
    pass


class ConfigurationError(Exception):
    pass


class CreateS3Account(Step):
    def __init__(self, name: str, sudo_password: str, host: str):
        self.name = name
        self.sudo_password = sudo_password
        self.host = host

    def __str__(self) -> str:
        return f"Create S3 account 'RG {self.name}'"

    def verify(self, _):
        if rg_account_exists(self.name, self.host, self.sudo_password):
            raise StateAlreadyExists(f"S3 account {self.name} already exists!")

    def apply(self, _):
        remote_sudo(
            self.host,
            self.sudo_password,
            [
                "radosgw-admin",
                "user",
                "create",
                "--uid",
                self.name,
                "--display-name",
                f"RG {self.name}",
            ],
        )

        return True


class Keypair:
    def __init__(self, key_id: str, secret_key: str):
        self.key_id = key_id
        self.secret_key = secret_key


class CreateKeypair(Step):
    def __init__(self, rg_name: str, sudo_password: str, host: str):
        self.rg_name = rg_name
        self.sudo_password = sudo_password
        self.host = host

    def __str__(self) -> str:
        return f"Generate keypair for account 'RG {self.rg_name}'"

    def verify(self, account_created):
        if not account_created and not rg_account_exists(
            self.rg_name, self.host, self.sudo_password
        ):
            raise ConfigurationError(
                f"Account for rg {self.rg_name} does not exist!"
            )

    def apply(self, _) -> Keypair:
        data = json.loads(
            remote_sudo(
                self.host,
                self.sudo_password,
                [
                    "radosgw-admin",
                    "key",
                    "create",
                    "--uid",
                    self.rg_name,
                    "--gen-access-key",
                ],
            )
        )
        kp = Keypair(
            data["keys"][-1]["access_key"], data["keys"][-1]["secret_key"]
        )

        print(f"AWS_ACCESS_KEY_ID={kp.key_id}")
        print(f"AWS_SECRET_ACCESS_KEY={kp.secret_key}")

        return kp


class CreateBucket(Step):
    def __init__(self, name: str):
        self.name = name

    def __str__(self) -> str:
        return f"Create bucket {self.name}"

    def verify(self, key: Union[Keypair, None]):
        print_bold(f"Creating bucket {self.name}...")
        kwargs = {"endpoint_url": "https://files.flyingcircus.io/"}
        if key is not None:
            kwargs["aws_access_key_id"] = key.key_id
            kwargs["aws_secret_access_key"] = key.secret_key
        self.resource = boto3.resource("s3", **kwargs)
        self.client = boto3.client("s3", **kwargs)

        self.bucket = self.resource.Bucket(self.name)
        if self.bucket.creation_date:
            raise StateAlreadyExists(
                f"Bucket with name {self.name} already exists!"
            )

    def apply(self, keys):
        self.bucket.create()
        return self.client


class Rules:
    rules: Dict[str, int] = {}

    def __iter__(self):
        return iter(self.rules.items())

    def __len__(self):
        return len(self.rules)

    def add(self, prefix: str, days: int):
        if prefix in self.rules:
            raise KeyError(f"Key '{prefix}' already has a rule")
        self.rules[prefix] = days

    def __str__(self) -> str:
        return ", ".join(
            f"'{prefix}' expires in {days} days" for prefix, days in self
        )


class CreateLifecyclePolicyConfiguration(Step):
    def __init__(self, bucket: str, rules: Rules):
        self.bucket = bucket
        self.rules = rules

    def __str__(self) -> str:
        return f"Create {len(self.rules)} lifecycle configuration(s) for s3://{self.bucket} ({self.rules})"

    def apply(self, client):
        print_bold("Creating lifecycle rules...")
        client.put_bucket_lifecycle_configuration(
            Bucket=self.bucket,
            LifecycleConfiguration={
                "Rules": [
                    {
                        "Expiration": {"Days": days},
                        "Prefix": prefix,
                        "Status": "Enabled",
                    }
                    for prefix, days in self.rules
                ]
            },
        )


def ask_sudo_password() -> str:
    return inquirer.secret(
        message="What's your FCIO password (needed to run radosgw-admin)?"
    ).execute()


def run():
    print_bold(
        "Interactively creating an S3 bucket + keys as described in https://wiki.flyingcircus.io/S3"
    )

    plan = Plan()
    sudo_password = None
    ceph_osd_host = None
    rg_name = inquirer.text(message="What's the name of the RG?").execute()
    keypair_needed = False
    if inquirer.confirm(message="Does the RG need to be created?").execute():
        keypair_needed = True
        sudo_password = ask_sudo_password()
        ceph_osd_host = inquirer.text(
            message="Which host to use to call radosgw-admin (must be an FQDN)?"
        ).execute()
        plan += CreateS3Account(rg_name, sudo_password, ceph_osd_host)

    if (
        keypair_needed
        or inquirer.confirm(
            message="Do you need a keypair for access?"
        ).execute()
    ):
        if sudo_password is None:
            sudo_password = ask_sudo_password()
        if ceph_osd_host is None:
            ceph_osd_host = inquirer.text(
                message="Which host to use to call radosgw-admin (must be an FQDN)?"
            ).execute()
        plan += CreateKeypair(rg_name, sudo_password, ceph_osd_host)
    elif not all(
        x in environ for x in ["AWS_ACCESS_KEY_ID", "AWS_SECRET_ACCESS_KEY"]
    ):
        error(
            "AWS_ACCESS_KEY_ID/AWS_SECRET_ACCESS_KEY missing from environment"
        )
        exit(1)

    bucket_name = inquirer.text(
        message="What's the name of the bucket?"
    ).execute()
    plan += CreateBucket(bucket_name)

    rules = Rules()
    while inquirer.confirm(
        message="Do you want to create a bucket lifecycle configuration?"
    ).execute():
        try:
            rules.add(
                inquirer.text(
                    message="For which prefix should this policy apply (leading slash will be stripped)?"
                )
                .execute()
                .lstrip("/"),
                int(
                    inquirer.number(
                        message="After how many days should the objects be expired?"
                    ).execute()
                ),
            )
        except KeyError as e:
            print(e)

    if len(rules) > 0:
        plan += CreateLifecyclePolicyConfiguration(bucket_name, rules)

    print_bold("Would perform the following steps:")
    print(plan)

    if inquirer.confirm(message="Does this look OK?").execute():
        plan.apply()
    else:
        print("Abort. Nothing changed.")


def main():
    try:
        run()
    except KeyboardInterrupt:
        print("Interrupt")
        exit(130)
    except EOFError:
        print("Aborted")
        exit(130)


if __name__ == "__main__":
    main()



import argparse
import configparser
import json
import os
import os.path
import subprocess
import sys

import batou.environment
from InquirerPy import inquirer
from prompt_toolkit.completion import FuzzyWordCompleter, ThreadedCompleter

from batou_ext.jenkins import VersionsUpdater


def get_git_version_completer(url):
    def get_words():
        cmd = subprocess.Popen(
            ["git", "ls-remote", url], stdout=subprocess.PIPE, encoding="UTF-8"
        )
        stdout, stderr = cmd.communicate()
        words = []
        for line in stdout.splitlines():
            words.extend(line.split())
        return words

    return ThreadedCompleter(FuzzyWordCompleter(get_words))


class Updater:
    """Version update, interactive or automatic."""

    environment_name: str = None
    _environment = None
    _target_versions = None

    def __init__(self, basedir):
        self.basedir = basedir

    @property
    def environment(self):
        if self._environment is None:
            assert self.environment_name
            self._environment = batou.environment.Environment(
                self.environment_name
            )
            self._environment.load()
        return self._environment

    def update_from_branch(self, branch: str):
        envs = sorted(os.listdir(os.path.join(self.basedir, "environments")))
        for env_name in envs:
            env = batou.environment.Environment(env_name)
            env.load()
            if env.branch == branch:
                print(f"Updating environment: {env_name}")
                self.environment_name = env_name
                self.set_versions()

        if not self.environment_name:
            raise ValueError(
                f"Branch {branch} is not used in any environment, "
                "cannot auto-upate."
            )

    @property
    def versions_ini(self):
        if "versions_ini" in self.environment.overrides["settings"]:
            return self.environment.overrides["settings"]["versions_ini"]
        elif os.path.exists("versions.ini"):
            return "versions.ini"
        else:
            raise ValueError(
                "No versions.ini specified in the environment file (via `settings.versions_ini`) and the file `versions.ini` does not exist in the git root, cannot proceed"
            )

    def interactive(self):
        envs = os.listdir(os.path.join(self.basedir, "environments"))
        envs.sort()
        self.environment_name = inquirer.select(
            message="Select environment to update",
            choices=envs,
        ).execute()

        versions = configparser.ConfigParser()
        versions.read(os.path.join(self.basedir, self.versions_ini))
        components = versions.sections()
        selected_components = inquirer.select(
            message="Select components to update (Space to select, Ctrl-A for all):",
            choices=components,
            multiselect=True,
        ).execute()

        self._target_versions = {}
        for c in selected_components:
            try:
                default = versions.get(c, "default")
            except configparser.NoOptionError:
                default = ""

            git_url = versions[c].get("url")
            if git_url:
                git_completer = get_git_version_completer(git_url)
            else:
                git_completer = None
            new_version = inquirer.text(
                message=f"Update {c} to:",
                default=default,
                completer=git_completer,
            ).execute()

            self._target_versions[c] = new_version

    def set_versions(self):
        if self._target_versions is None:
            self._target_versions = {}
            versions = configparser.ConfigParser()
            versions.read(os.path.join(self.basedir, self.versions_ini))
            for section in versions.sections():
                try:
                    self._target_versions[section] = versions.get(
                        section, "default"
                    )
                except configparser.NoOptionError:
                    pass

        updater = VersionsUpdater(
            self.versions_ini, json.dumps(self._target_versions)
        )
        updater()
        subprocess.run(["git", "-P", "diff", self.versions_ini])


def main():
    basedir = os.environ.get("APPENV_BASEDIR", os.path.dirname(sys.argv[0]))

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-a",
        "--update-defaults",
        dest="environment",
        help="Auto-update all versions to their specified defaults.",
    )
    parser.add_argument(
        "-b",
        "--branch",
        dest="branch",
        help="Update environment matching the branch.",
    )

    parser.add_argument(
        "-d",
        "--base-dir",
        dest="basedir",
        help="Deployment base dir",
        default=basedir,
    )

    args = parser.parse_args()

    update = Updater(args.basedir)

    if args.environment:
        update.environment_name = args.environment
        update.set_versions()
    elif args.branch:
        update.update_from_branch(args.branch)
    else:
        update.interactive()
        update.set_versions()


if __name__ == "__main__":
    main()
